@@ -1,12 +1,16 @@
+#! /usr/bin/perl -w
+
 package PostScript::Simple;
 
 use strict;
 use vars qw($VERSION @ISA @EXPORT);
+use Carp;
 use Exporter;
+use PostScript::Simple::EPS;
 
 @ISA = qw(Exporter);
 @EXPORT = qw();
-$VERSION = '0.05';
+$VERSION = '0.06p3';
 
 =head1 NAME
 
@@ -19,6 +23,7 @@ PostScript::Simple - Produce PostScript files from Perl
     # create a new PostScript object
     $p = new PostScript::Simple(papersize => "A4",
                                 colour => 1,
+                                eps => 0,
                                 units => "in");
     
     # create a new page
@@ -233,7 +238,12 @@ remember the size of paper using PostScript points. Valid choices are currently
 =item landscape
 
 Use the landscape option to rotate the page by 90 degrees. The paper dimensions
-are also rotated, so that clipping will still work. (Default: 0)
+are also rotated, so that clipping will still work. (Note that the printer will
+still think that the paper is portrait.) (Default: 0)
+
+=item copies
+
+Set the number of copies that should be printed. (Default: 1)
 
 =item clip
 
@@ -317,36 +327,41 @@ sub new# {{{
 {
   my ($class, %data) = @_;
   my $self = {
-    xsize        => undef,
-    ysize        => undef,
-    papersize    => undef,
-    units        => "bp",     # measuring units (see below)
-    landscape    => 0,        # rotate the page 90 degrees
-    colour       => 0,        # use colour
-    clip         => 0,        # clip to the bounding box
-    eps          => 1,        # create eps file
-    page         => 1,        # page number to start at
-    reencode     => "ISOLatin1Encoding", # Re-encode the 13 std fonts in this encoding
+    xsize          => undef,
+    ysize          => undef,
+    papersize      => undef,
+    units          => "bp",     # measuring units (see below)
+    landscape      => 0,        # rotate the page 90 degrees
+    copies         => 1,        # number of copies
+    colour         => 0,        # use colour
+    clip           => 0,        # clip to the bounding box
+    eps            => 1,        # create eps file
+    page           => 1,        # page number to start at
+    reencode       => "ISOLatin1Encoding", # Re-encode the 13 standard
+                                           # fonts in this encoding
 
-    bbx1         => 0,        # Bounding Box definitions
-    bby1         => 0,
-    bbx2         => 0,
-    bby2         => 0,
+    bbx1           => 0,        # Bounding Box definitions
+    bby1           => 0,
+    bbx2           => 0,
+    bby2           => 0,
 
-    pscomments   => "",       # the following entries store data
-    psprolog     => "",       # for the same DSC areas of the
-    psfunctions  => "",       # postscript file.
-    pssetup      => "",
-    pspages      => "",
-    pstrailer    => "",
+    pscomments     => "",       # the following entries store data
+    psprolog       => "",       # for the same DSC areas of the
+    psfunctions    => "",       # postscript file.
+    pssetup        => "",
+    pspages        => "",
+    pstrailer      => "",
 
-    pspagecount  => 0,
-    usedcircle   => 0,
-    usedbox      => 0,
-    usedrotabout => 0,
+    lastfontsize   => 0,
+    pspagecount    => 0,
+    usedcircle     => 0,
+    usedcircletext => 0,
+    usedbox        => 0,
+    usedrotabout   => 0,
+    usedimporteps  => 0,
 
-    coordorigin  => 'LeftBottom',
-    direction    => 'RightUp',
+    coordorigin    => 'LeftBottom',
+    direction      => 'RightUp',
   };
 
   foreach (keys %data)
@@ -457,7 +472,10 @@ sub init# {{{
   90 rotate
 } bind def
 ";
-    $self->{pscomments} .= "\%\%Orientation: Landscape\n";
+    # I now think that Portrait is the correct thing here, as the page is
+    # rotated.
+    $self->{pscomments} .= "\%\%Orientation: Portrait\n";
+#    $self->{pscomments} .= "\%\%Orientation: Landscape\n";
     $swap = $self->{bbx2};
     $self->{bbx2} = $self->{bby2};
     $self->{bby2} = $swap;
@@ -680,89 +698,172 @@ sub newpage# {{{
 Writes the current PostScript out to the file named C<filename>. Will destroy
 any existing file of the same name.
 
-Use this option whenever output is required to disk. The current PostScript
+Use this method whenever output is required to disk. The current PostScript
 document in memory is not cleared, and can still be extended.
 
 =cut
 
 
-sub output# {{{
+sub _buildpage# {{{
 {
   my $self = shift;
-  my $file = shift || die("Must supply a filename for output");
+  my $title = shift;
   
-  my $eps = $self->{eps};
+  my $page;
   my $date = scalar localtime;
   my $user;
 
+  $title = 'undefined' unless $title;
+
+  $page = [];
+
 # getlogin is unimplemented on some systems
   eval { $user = getlogin; };
   $user = 'Console' unless $user;
 
-  local *OUT;
-  open(OUT, '>'.$file) or die("Cannot write to file $file: $!");
-  select OUT;
-
 # Comments Section
-  print "%!PS-Adobe-3.0";
-  if ($eps) { print " EPSF-1.2" }
-  print "\n";
-  print "\%\%Title: ($file)\n";
-  print "\%\%LanguageLevel: 1\n";
-  print "\%\%Creator: PostScript::Simple perl module version $VERSION\n";
-  print "\%\%CreationDate: $date\n";
-  print "\%\%For: $user\n";
-  print $self->{pscomments};
-#  print "\%\%DocumentFonts: \n";
-  if ($eps)
+  push @$page, "%!PS-Adobe-3.0";
+  push @$page, " EPSF-1.2" if ($self->{eps});
+  push @$page, "\n";
+  push @$page, "\%\%Title: ($title)\n";
+  push @$page, "\%\%LanguageLevel: 1\n";
+  push @$page, "\%\%Creator: PostScript::Simple perl module version $VERSION\n";
+  push @$page, "\%\%CreationDate: $date\n";
+  push @$page, "\%\%For: $user\n";
+  push @$page, \$self->{pscomments};
+#  push @$page, "\%\%DocumentFonts: \n";
+  if ($self->{eps})
   {
-    print "\%\%BoundingBox: $self->{bbx1} $self->{bby1} $self->{bbx2} $self->{bby2}\n";
+    push @$page, "\%\%BoundingBox: $self->{bbx1} $self->{bby1} $self->{bbx2} $self->{bby2}\n";
   }
   else
   {
-    print "\%\%Pages: $self->{pspagecount}\n";
+    push @$page, "\%\%Pages: $self->{pspagecount}\n";
   }
-  print "\%\%EndComments\n";
+  push @$page, "\%\%EndComments\n";
   
 # Prolog Section
-  print "\%\%BeginProlog\n";
-  print $self->{psprolog};
-  print "\%\%BeginResource: PostScript::Simple\n";
-  print $self->{psfunctions};
-  print "\%\%EndResource\n";
-  print "\%\%EndProlog\n";
+  push @$page, "\%\%BeginProlog\n";
+  push @$page, \$self->{psprolog};
+  push @$page, "\%\%BeginResource: PostScript::Simple\n";
+  push @$page, \$self->{psfunctions};
+  push @$page, "\%\%EndResource\n";
+  push @$page, "\%\%EndProlog\n";
 
 # Setup Section
-  if (length($self->{pssetup}))
+  if (length($self->{pssetup}) || ($self->{copies} > 1))
   {
-    print "\%\%BeginSetup\n";
-    print $self->{pssetup};
-    print "\%\%EndSetup\n";
+    push @$page, "\%\%BeginSetup\n";
+    if ($self->{copies} > 1)
+    {
+      push @$page, "/#copies " . $self->{copies} . " def\n";
+    }
+    push @$page, \$self->{pssetup};
+    push @$page, "\%\%EndSetup\n";
   }
 
 # Pages
-  print $self->{pspages};
-  if ((!$eps) && ($self->{pspagecount} > 0))
+  push @$page, \$self->{pspages};
+  if ((!$self->{eps}) && ($self->{pspagecount} > 0))
   {
-    print "\%\%PageTrailer\n";
-    print "pagelevel restore\n";
-    print "showpage\n";
+    push @$page, "\%\%PageTrailer\n";
+    push @$page, "pagelevel restore\n";
+    push @$page, "showpage\n";
   }
 
 # Trailer Section
   if (length($self->{pstrailer}))
   {
-    print "\%\%Trailer\n";
-    print $self->{pstrailer};
+    push @$page, "\%\%Trailer\n";
+    push @$page, \$self->{pstrailer};
   }
-  print "\%\%EOF\n";
-  select STDOUT;
+  push @$page, "\%\%EOF\n";
+  
+  return $page;
+}# }}}
+
+sub output# {{{
+{
+  my $self = shift;
+  my $file = shift || die("Must supply a filename for output");
+  my $page;
+  my $i;
+  
+  $page = _buildpage($self, $file);
+
+  local *OUT;
+  open(OUT, '>'.$file) or die("Cannot write to file $file: $!");
+
+  foreach $i (@$page) {
+    if (ref($i) eq "SCALAR") {
+      print OUT $$i;
+    } else {
+      print OUT $i;
+    }
+  }
+
   close OUT;
   
   return 1;
 }# }}}
 
 
+=item C<get>
+
+Returns the current document.
+
+Use this method whenever output is required as a scalar. The current PostScript
+document in memory is not cleared, and can still be extended.
+
+=cut
+
+sub get# {{{
+{
+  my $self = shift;
+  my $page;
+  my $i;
+  my $doc;
+  
+  $page = _buildpage($self, "PostScript::Simple generated page");
+  $doc = "";
+  foreach $i (@$page) {
+    if (ref($i) eq "SCALAR") {
+      $doc .= $$i;
+    } else {
+      $doc .= $i;
+    }
+  }
+  return $doc;
+}# }}}
+
+
+=item C<geteps>
+
+Returns the current document as a PostScript::Simple::EPS object. Only works if
+the current document is EPS.
+
+This method calls new PostScript::Simple::EPS with all the default options. To
+change these, call it yourself as below, rather than using this method.
+
+  $eps = new PostScript::Simple::EPS(source => $ps->get);
+
+=cut
+
+sub geteps# {{{
+{
+  my $self = shift;
+  my $page;
+  my $i;
+  my $doc;
+  my $eps;
+  
+  croak "document is not EPS" unless ($$self{eps} == 1);
+
+  $eps = new PostScript::Simple::EPS(source => $self->get);
+  return $eps;
+}# }}}
+
+
 =item C<setcolour((red, green, blue)|(name))>
 
 Sets the new drawing colour to the values specified in C<red>, C<green> and
@@ -952,6 +1053,63 @@ sub linextend# {{{
   return 1;
 }# }}}
 
+=item C<arc([options,] x,y, radius, start_angle, end_angle)>
+
+Draws an arc on the circle of radius C<radius> with centre (C<x>,C<y>). The arc
+starts at angle C<start_angle> and finishes at C<end_angle>. Angles are specified
+in degrees, where 0 is at 3 o'clock, and the direction of travel is anti-clockwise.
+
+Any options are passed in a hash reference as the first parameter. The available
+option is:
+
+=over 4
+
+=item filled => 1
+
+If C<filled> is 1 then the arc will be filled in.
+
+=back
+
+Example:
+
+    # semi-circle
+    $p->arc(10, 10, 5, 0, 180);
+
+    # complete filled circle
+    $p->arc({filled=>1}, 30, 30, 10, 0, 360);
+
+=cut
+
+sub arc# {{{
+{
+  my $self = shift;
+  my %opt = ();
+
+  if (ref($_[0])) {
+    %opt = %{; shift};
+  }
+
+  if ((!$self->{pspagecount}) and (!$self->{eps})) {
+# Cannot draw on to non-page when not an eps file XXXXX
+    return 0;
+  }
+
+  my ($x, $y, $r, $sa, $ea) = @_;
+
+  unless (@_ == 5) {
+    $self->_error("arc: wrong number of arguments");
+    return 0;
+  }
+
+  $self->{pspages} .= "$x ux $y uy $r u $sa $ea arc ";
+  if ($opt{'filled'}) {
+    $self->{pspages} .= "fill\n"
+  } else {
+    $self->{pspages} .= "stroke\n"
+  }
+  
+  return 1;
+}# }}}
 
 =item C<polygon([options,] x1,y1, x2,y2, ..., xn,yn)>
 
@@ -1175,6 +1333,123 @@ sub circle# {{{
   return 1;
 }# }}}
 
+=item C<circletext([options,] x, y, r, a, text)>
+
+Draw text in an arc centered about angle C<a> with circle midpoint (C<x>,C<y>)
+and radius C<r>.
+
+There is only one option.
+
+=over 4
+
+=item align => "alignment"
+
+C<alignment> can be 'inside' or 'outside'. The default is 'inside'.
+
+=back
+
+Example:
+
+    # outside the radius, centered at 90 degrees from the origin
+    $p->circletext(40, 40, 20, 90, "Hello, Outside World!");
+    # inside the radius centered at 270 degrees from the origin
+    $p->circletext( {align => "inside"}, 40, 40, 20, 270, "Hello, Inside World!");
+
+=cut
+
+
+sub circletext# {{{
+{
+  my $self = shift;
+  my %opt = ();
+
+  if (ref($_[0]))
+  {
+    %opt = %{; shift};
+  }
+
+  my ($x, $y, $r, $a, $text) = @_;
+
+  unless (@_ == 5) {
+    $self->_error("circletext: wrong number of arguments");
+    return 0;
+  }
+
+  unless (defined $self->{lastfontsize}) {
+    $self->_error("circletext: must set font first");
+    return 0;
+  }
+
+  if (!$self->{usedcircletext}) {
+    $self->{psfunctions} .= <<'EOCT';
+/outsidecircletext
+  { $circtextdict begin
+      /radius exch def
+      /centerangle exch def
+      /ptsize exch def
+      /str exch def
+      /xradius radius ptsize 4 div add def
+      gsave
+        centerangle str findhalfangle add rotate
+        str { /charcode exch def ( ) dup 0 charcode put outsideshowcharandrotate } forall
+      grestore
+    end
+  } def
+       
+/insidecircletext
+  { $circtextdict begin
+      /radius exch def
+      /centerangle exch def
+      /ptsize exch def
+      /str exch def
+      /xradius radius ptsize 3 div sub def
+      gsave
+        centerangle str findhalfangle sub rotate
+        str { /charcode exch def ( ) dup 0 charcode put insideshowcharandrotate } forall
+      grestore
+    end
+  } def
+/$circtextdict 16 dict def
+$circtextdict begin
+  /findhalfangle
+    { stringwidth pop 2 div 2 xradius mul pi mul div 360 mul
+    } def
+  /outsideshowcharandrotate
+    { /char exch def
+      /halfangle char findhalfangle def
+      gsave
+        halfangle neg rotate radius 0 translate -90 rotate
+        char stringwidth pop 2 div neg 0 moveto char show
+      grestore
+      halfangle 2 mul neg rotate
+    } def
+  /insideshowcharandrotate
+    { /char exch def
+      /halfangle char findhalfangle def
+      gsave
+        halfangle rotate radius 0 translate 90 rotate
+        char stringwidth pop 2 div neg 0 moveto char show
+      grestore
+      halfangle 2 mul rotate
+    } def
+  /pi 3.1415926 def
+end
+EOCT
+    $self->{usedcircletext} = 1;
+  }
+
+  $self->{pspages} .= "gsave\n";
+  $self->{pspages} .= "  $x ux $y uy translate\n";
+  $self->{pspages} .= "  ($text) $self->{lastfontsize} $a $r u ";
+  if ($opt{'align'} && ($opt{'align'} eq "outside")) {
+    $self->{pspages} .= "outsidecircletext\n";
+  } else {
+    $self->{pspages} .= "insidecircletext\n";
+  }
+  $self->{pspages} .= "grestore\n";
+  
+  return 1;
+}# }}}
 
 =item C<box(x1,y1, x2,y2 [, options])>
 
@@ -1271,6 +1546,8 @@ sub setfont# {{{
 # set font y size XXXXX
   $self->{pspages} .= "/$name findfont $size scalefont setfont\n";
 
+  $self->{lastfontsize} = $size;
+
   return 1;
 }# }}}
 
@@ -1465,6 +1742,213 @@ sub moveto# {{{
 }# }}}
 
 
+=item C<importepsfile([options,] filename, x1,y1, x2,y2)>
+
+Imports an EPS file and scales/translates its bounding box to fill
+the area defined by lower left co-ordinates (x1,y1) and upper right
+co-ordinates (x2,y2). By default, if the co-ordinates have a different
+aspect ratio from the bounding box, the scaling is constrained on the
+greater dimension to keep the EPS fully inside the area.
+
+Options are:
+
+=over 4
+
+=item overlap => 1
+
+If C<overlap> is 1 then the scaling is calculated on the lesser dimension
+and the EPS can overlap the area.
+
+=item stretch => 1
+
+If C<stretch> is 1 then fill the entire area, ignoring the aspect ratio.
+This option overrides C<overlap> if both are given.
+
+=back
+
+Example:
+
+    # Assume smiley.eps is a round smiley face in a square bounding box
+
+    # Scale it to a (10,10)(20,20) box
+    $p->importepsfile("smiley.eps", 10,10, 20,20);
+
+    # Keeps aspect ratio, constrained to smallest fit
+    $p->importepsfile("smiley.eps", 10,10, 30,20);
+
+    # Keeps aspect ratio, allowed to overlap for largest fit
+    $p->importepsfile( {overlap => 1}, "smiley.eps", 10,10, 30,20);
+
+    # Aspect ratio is changed to give exact fit
+    $p->importepsfile( {stretch => 1}, "smiley.eps", 10,10, 30,20);
+
+=cut
+
+
+sub importepsfile# {{{
+{
+  my $self = shift;
+
+  my $bbllx;
+  my $bblly;
+  my $bburx;
+  my $bbury;
+  my $bbw;
+  my $bbh;
+  my $pagew;
+  my $pageh;
+  my $scalex;
+  my $scaley;
+  my $line;
+  my $eps;
+
+  my %opt = ();
+
+  if (ref($_[0])) {
+    %opt = %{; shift};
+  }
+
+  my ($file, $x1, $y1, $x2, $y2) = @_;
+
+  unless (@_ == 5) {
+    $self->_error("importepsfile: wrong number of arguments");
+    return 0;
+  }
+
+  $opt{'overlap'} = 0 if (!defined($opt{'overlap'}));
+  $opt{'stretch'} = 0 if (!defined($opt{'stretch'}));
+  
+  $eps = new PostScript::Simple::EPS(file => $file);
+  ($bbllx, $bblly, $bburx, $bbury) = $eps->get_bbox();
+
+  $pagew = $x2 - $x1;
+  $pageh = $y2 - $y1;
+
+  $bbw = $bburx - $bbllx;
+  $bbh = $bbury - $bblly;
+
+  if (($bbw == 0) || ($bbh == 0)) {
+    $self->_error("importeps: Bounding Box has zero dimension");
+    return 0;
+  }
+
+  $scalex = $pagew / $bbw;
+  $scaley = $pageh / $bbh;
+
+  if ($opt{'stretch'} == 0) {
+    if ($opt{'overlap'} == 0) {
+      if ($scalex > $scaley) {
+        $scalex = $scaley;
+      } else {
+        $scaley = $scalex;
+      }
+    } else {
+      if ($scalex > $scaley) {
+        $scaley = $scalex;
+      } else {
+        $scalex = $scaley;
+      }
+    }
+  }
+
+  $eps->scale($scalex, $scaley);
+  $eps->translate(-$bbllx, -$bblly);
+  $self->_add_eps($eps, $x1, $y1);
+
+  return 1;
+}# }}}
+
+
+=item C<importeps(filename, x,y)>
+
+Imports a PostScript::Simple::EPS object into the current document at position
+C<(x,y)>.
+
+Example:
+
+    use PostScript::Simple;
+    
+    # create a new PostScript object
+    $p = new PostScript::Simple(papersize => "A4",
+                                colour => 1,
+                                units => "in");
+    
+    # create a new page
+    $p->newpage;
+    
+    # create an eps object
+    $e = new PostScript::Simple::EPS(file => "test.eps");
+    $e->rotate(90);
+    $e->scale(0.5);
+
+    # add eps to the current page
+    $p->importeps($e, 10,50);
+
+=cut
+
+
+sub importeps# {{{
+{
+  my $self = shift;
+  my ($epsobj, $xpos, $ypos) = @_;
+
+  unless (@_ == 3) {
+    $self->_error("importeps: wrong number of arguments");
+    return 0;
+  }
+
+  $self->_add_eps($epsobj, $xpos, $ypos);
+
+  return 1;
+}# }}}
+
+sub _add_eps# {{{
+{
+  my $self = shift;
+  my $epsobj;
+  my $xpos;
+  my $ypos;
+
+  if (ref($_[0]) ne "PostScript::Simple::EPS") {
+    croak "internal error: _add_eps[0] must be eps object";
+  }
+
+  if ((!$self->{pspagecount}) and (!$self->{eps})) {
+    # Cannot draw on to non-page when not an eps file
+    $self->_error("importeps: no current page");
+    return 0;
+  }
+
+  if ( @_ != 3 ) {
+  	croak "internal error: wrong number of arguments for _add_eps";
+  	return 0;
+  }
+
+  unless ($self->{usedimporteps}) {
+    $self->{psfunctions} .= <<'EOEPS';
+/BeginEPSF { /b4_Inc_state save def /dict_count countdictstack def
+/op_count count 1 sub def userdict begin /showpage { } def 0 setgray
+0 setlinecap 1 setlinewidth 0 setlinejoin 10 setmiterlimit [ ]
+0 setdash newpath /languagelevel where { pop languagelevel 1 ne {
+false setstrokeadjust false setoverprint } if } if } bind def
+/EndEPSF { count op_count sub {pop} repeat countdictstack dict_count
+sub {end} repeat b4_Inc_state restore } bind def
+EOEPS
+    $self->{usedimporteps} = 1;
+  }
+
+  ($epsobj, $xpos, $ypos) = @_;
+
+  $self->{pspages} .= "BeginEPSF\n";
+  $self->{pspages} .= "$xpos ux $ypos uy translate\n";
+  $self->{pspages} .= "1 ux 1 uy scale\n";
+  $self->{pspages} .= $epsobj->_get_include_data($xpos, $ypos);
+  $self->{pspages} .= "EndEPSF\n";
+  
+  return 1;
+}# }}}
+
+
 ### PRIVATE
 
 sub _error {# {{{
@@ -1491,23 +1975,23 @@ sub _error {# {{{
 =head1 BUGS
 
 Some current functionality may not be as expected, and/or may not work correctly.
-
-The way of giving options to methods has changed since version 0.04, this may break
-your code; sorry.
-
-More functions need to be added. See the README file.
+That's the fun with using code in development!
 
 =head1 AUTHOR
 
-The PostScript::Simple module was initially created by Matthew Newton, with
-ideas and suggestions from Mark Withall.
+The PostScript::Simple module was created by Matthew Newton, with ideas
+and suggestions from Mark Withall.
 
 Contributions to the work (either code or suggestions / improvements) have been
 gratefully received from: Andreas Marcel, P Kent, Flemming Frandsen, Michael
-Tomuschat, Vladi Belperchinov-Shabanski
+Tomuschat, Vladi Belperchinov-Shabanski, Eric Wilhelm, Peter Kuhn, Glen Harris.
+
+Apologies if you have contributed and I have missed you... please let me know so
+I can add your name to the next release.
 
 =head1 SEE ALSO
 
+L<PostScript::Simple::EPS>
 L<GD>
 
 =cut
